@@ -12,14 +12,31 @@ import sys
 import threading
 
 from types import FrameType, ModuleType
-from typing import Any, Callable, Dict, List, Optional, Set, Tuple, cast
+from typing import Any, Callable, cast
 
 from coverage import env
 from coverage.types import (
-    TArc, TFileDisposition, TLineNo, TTraceData, TTraceFileData, TTraceFn,
-    TracerCore, TWarnFn,
+    TArc,
+    TFileDisposition,
+    TLineNo,
+    TShouldStartContextFn,
+    TShouldTraceFn,
+    TTraceData,
+    TTraceFileData,
+    TTraceFn,
+    TWarnFn,
+    Tracer,
 )
 
+
+# I don't understand why, but if we use `cast(set[TLineNo], ...)` inside
+# the _trace() function, we get some strange behavior on PyPy 3.10.
+# Assigning these names here and using them below fixes the problem.
+# See https://github.com/nedbat/coveragepy/issues/1902
+set_TLineNo = set[TLineNo]
+set_TArc = set[TArc]
+
+
 # We need the YIELD_VALUE opcode below, in a comparison-friendly form.
 # PYVERSIONS: RESUME is new in Python3.11
 RESUME = dis.opmap.get("RESUME")
@@ -28,13 +45,15 @@ if RESUME is None:
     YIELD_VALUE = dis.opmap["YIELD_VALUE"]
     YIELD_FROM = dis.opmap["YIELD_FROM"]
     YIELD_FROM_OFFSET = 0 if env.PYPY else 2
+else:
+    YIELD_VALUE = YIELD_FROM = YIELD_FROM_OFFSET = -1
 
 # When running meta-coverage, this file can try to trace itself, which confuses
 # everything.  Don't trace ourselves.
 
 THIS_FILE = __file__.rstrip("co")
 
-class PyTracer(TracerCore):
+class PyTracer(Tracer):
     """Python implementation of the raw data tracer."""
 
     # Because of poor implementations of trace-function-manipulating tools,
@@ -62,19 +81,21 @@ class PyTracer(TracerCore):
         # Attributes set from the collector:
         self.data: TTraceData
         self.trace_arcs = False
-        self.should_trace: Callable[[str, FrameType], TFileDisposition]
-        self.should_trace_cache: Dict[str, Optional[TFileDisposition]]
-        self.should_start_context: Optional[Callable[[FrameType], Optional[str]]] = None
-        self.switch_context: Optional[Callable[[Optional[str]], None]] = None
+        self.should_trace: TShouldTraceFn
+        self.should_trace_cache: dict[str, TFileDisposition | None]
+        self.should_start_context: TShouldStartContextFn | None = None
+        self.switch_context: Callable[[str | None], None] | None = None
+        self.lock_data: Callable[[], None]
+        self.unlock_data: Callable[[], None]
         self.warn: TWarnFn
 
         # The threading module to use, if any.
-        self.threading: Optional[ModuleType] = None
+        self.threading: ModuleType | None = None
 
-        self.cur_file_data: Optional[TTraceFileData] = None
+        self.cur_file_data: TTraceFileData | None = None
         self.last_line: TLineNo = 0
-        self.cur_file_name: Optional[str] = None
-        self.context: Optional[str] = None
+        self.cur_file_name: str | None = None
+        self.context: str | None = None
         self.started_context = False
 
         # The data_stack parallels the Python call stack. Each entry is
@@ -85,8 +106,8 @@ class PyTracer(TracerCore):
         #           this frame.
         #   [2] The last line number executed in this frame.
         #   [3] Boolean: did this frame start a new context?
-        self.data_stack: List[Tuple[Optional[TTraceFileData], Optional[str], TLineNo, bool]] = []
-        self.thread: Optional[threading.Thread] = None
+        self.data_stack: list[tuple[TTraceFileData | None, str | None, TLineNo, bool]] = []
+        self.thread: threading.Thread | None = None
         self.stopped = False
         self._activity = False
 
@@ -106,11 +127,7 @@ class PyTracer(TracerCore):
     def log(self, marker: str, *args: Any) -> None:
         """For hard-core logging of what this tracer is doing."""
         with open("/tmp/debug_trace.txt", "a") as f:
-            f.write("{} {}[{}]".format(
-                marker,
-                self.id,
-                len(self.data_stack),
-            ))
+            f.write(f"{marker} {self.id}[{len(self.data_stack)}]")
             if 0:   # if you want thread ids..
                 f.write(".{:x}.{:x}".format(                    # type: ignore[unreachable]
                     self.thread.ident,
@@ -131,8 +148,8 @@ class PyTracer(TracerCore):
         frame: FrameType,
         event: str,
         arg: Any,                               # pylint: disable=unused-argument
-        lineno: Optional[TLineNo] = None,       # pylint: disable=unused-argument
-    ) -> Optional[TTraceFn]:
+        lineno: TLineNo | None = None,       # pylint: disable=unused-argument
+    ) -> TTraceFn | None:
         """The trace function passed to sys.settrace."""
 
         if THIS_FILE in frame.f_code.co_filename:
@@ -160,7 +177,7 @@ class PyTracer(TracerCore):
                     "Empty stack!",
                     frame.f_code.co_filename,
                     frame.f_lineno,
-                    frame.f_code.co_name
+                    frame.f_code.co_name,
                 )
             return None
 
@@ -170,12 +187,12 @@ class PyTracer(TracerCore):
         if event == "call":
             # Should we start a new context?
             if self.should_start_context and self.context is None:
-                context_maybe = self.should_start_context(frame)
+                context_maybe = self.should_start_context(frame)    # pylint: disable=not-callable
                 if context_maybe is not None:
                     self.context = context_maybe
                     started_context = True
                     assert self.switch_context is not None
-                    self.switch_context(self.context)
+                    self.switch_context(self.context)   # pylint: disable=not-callable
                 else:
                     started_context = False
             else:
@@ -190,7 +207,7 @@ class PyTracer(TracerCore):
                     self.cur_file_name,
                     self.last_line,
                     started_context,
-                )
+                ),
             )
 
             # Improve tracing performance: when calling a function, both caller
@@ -211,8 +228,12 @@ class PyTracer(TracerCore):
                 if disp.trace:
                     tracename = disp.source_filename
                     assert tracename is not None
-                    if tracename not in self.data:
-                        self.data[tracename] = set()
+                    self.lock_data()
+                    try:
+                        if tracename not in self.data:
+                            self.data[tracename] = set()
+                    finally:
+                        self.unlock_data()
                     self.cur_file_data = self.data[tracename]
                 else:
                     frame.f_trace_lines = False
@@ -241,9 +262,9 @@ class PyTracer(TracerCore):
                 flineno: TLineNo = frame.f_lineno
 
                 if self.trace_arcs:
-                    cast(Set[TArc], self.cur_file_data).add((self.last_line, flineno))
+                    cast(set_TArc, self.cur_file_data).add((self.last_line, flineno))
                 else:
-                    cast(Set[TLineNo], self.cur_file_data).add(flineno)
+                    cast(set_TLineNo, self.cur_file_data).add(flineno)
                 self.last_line = flineno
 
         elif event == "return":
@@ -274,7 +295,7 @@ class PyTracer(TracerCore):
                         real_return = True
                 if real_return:
                     first = frame.f_code.co_firstlineno
-                    cast(Set[TArc], self.cur_file_data).add((self.last_line, -first))
+                    cast(set_TArc, self.cur_file_data).add((self.last_line, -first))
 
             # Leaving this function, pop the filename stack.
             self.cur_file_data, self.cur_file_name, self.last_line, self.started_context = (
@@ -284,7 +305,7 @@ class PyTracer(TracerCore):
             if self.started_context:
                 assert self.switch_context is not None
                 self.context = None
-                self.switch_context(None)
+                self.switch_context(None)   # pylint: disable=not-callable
         return self._cached_bound_method_trace
 
     def start(self) -> TTraceFn:
@@ -345,6 +366,6 @@ class PyTracer(TracerCore):
         """Reset the activity() flag."""
         self._activity = False
 
-    def get_stats(self) -> Optional[Dict[str, int]]:
+    def get_stats(self) -> dict[str, int] | None:
         """Return a dictionary of statistics, or None."""
         return None
